// #####################################################################

// DS18B20 module

#include "ds18b20.h"
#include "menu.h"
 
// Data wire is plugged into pin 2 on the Arduino
#define ONE_WIRE_BUS1 14
#define ONE_WIRE_BUS2 13
 
// Setup a oneWire instance to communicate with any OneWire devices 
// (not just Maxim/Dallas temperature ICs)
OneWire oneWire1(ONE_WIRE_BUS1);
OneWire oneWire2(ONE_WIRE_BUS2);
 
// Pass our oneWire reference to Dallas Temperature.
DallasTemperature sensor1(&oneWire1);
DallasTemperature sensor2(&oneWire2);

float ds18b20_sensor1 = NAN;
float ds18b20_sensor2 = NAN;

char ds18b20_sensor1_txt[16];
char ds18b20_sensor2_txt[16];

char ds18b20_cycle = 0; // Cycle through sensors one at a time in global loop

struct MENU_ENTRY menu_inputs[] =
{          //  ################# - 17 char limit for 20x4 LCD
  { { .text = "DS18B20 Sensors" }, .type = MENU_LABEL, .data = NULL, .user_data = 0, .extra = NULL },
  { { .text = "DI #1: " }, .type = MENU_LABEL, .data = (void *) &ds18b20_sensor1_txt, .user_data = 0, .extra = NULL },
  { { .text = "DI #2: " }, .type = MENU_LABEL, .data = (void *) &ds18b20_sensor2_txt, .user_data = 0, .extra = NULL },
  { { .text = "" }, .type = MENU_NOTHING }
};

extern unsigned long current_millis;
unsigned long ds18b20_routine_millis = 5000;
bool b_ds18b20_initialized = false;

// #####################################################################

//
// Dallas Temperature libary non blocking question
//
// https://forum.arduino.cc/t/dallas-temperature-libary-non-blocking-question/153430/3
//

void ds18b20_routine( )
{
  // check timer
  if( current_millis < ds18b20_routine_millis )
    return;
  ds18b20_routine_millis = current_millis + 2500;

  if( b_ds18b20_initialized == false )
  {
    sensor1.begin();
    sensor2.begin();
    
    sensor1.setWaitForConversion( false );
    sensor2.setWaitForConversion( false );

    sensor1.requestTemperatures(); // Send the command to get temperatures
    sensor2.requestTemperatures(); // Send the command to get temperatures

    Serial.println(F("[DS18B20] External temperature subsystem initialized."));
    b_ds18b20_initialized = true;
  }
  else
  {
    //
    // Request temperature from only one sensor at a time in global loop
    //
    if( ds18b20_cycle == 0 )
    {
      ds18b20_sensor1 = sensor1.getTempCByIndex(0);
      sensor1.requestTemperatures(); // Send the command to get temperatures
      ds18b20_cycle++;
    }
    else if( ds18b20_cycle >= 1 )
    {
      ds18b20_sensor2 = sensor2.getTempCByIndex(0);
      sensor2.requestTemperatures(); // Send the command to get temperatures
      ds18b20_cycle = 0;
    }

    if( ds18b20_sensor1 == -127.0 ) ds18b20_sensor1 = NAN;
    if( ds18b20_sensor2 == -127.0 ) ds18b20_sensor2 = NAN;
    if( ds18b20_sensor1 == 85.0 ) ds18b20_sensor1 = NAN;
    if( ds18b20_sensor2 == 85.0 ) ds18b20_sensor2 = NAN;

    snprintf( ds18b20_sensor1_txt, sizeof(ds18b20_sensor1_txt), "%+5.1f " "\xDF" "C", ds18b20_sensor1 );
    snprintf( ds18b20_sensor2_txt, sizeof(ds18b20_sensor2_txt), "%+5.1f " "\xDF" "C", ds18b20_sensor2 );
  }
}

// #####################################################################