// #####################################################################

// General Purpose 16-bit Input/Output variables

#include "gpio.h"
#include "menu.h"
#include "keyboard.h"

PCF8575 PCF(0x20);

#define LED_CHARGING    9   // red
#define LED_TIMESYNC    10  // green
#define LED_WIFI        11  // blue
#define LED_IO          12  // red

#define GPIO_OUTPUT_1A_PIN    6
#define GPIO_OUTPUT_1B_PIN    7
#define GPIO_OUTPUT_2A_PIN    8
#define GPIO_OUTPUT_2B_PIN    13
#define GPIO_OUTPUT_3A_PIN    14
#define GPIO_OUTPUT_3B_PIN    15

struct GPIO_DIGITAL_OUTPUT gpio_digital_outputs[GPIO_DIGITAL_OUTPUTS] =
{
  { { .name = "DO #1A" }, .state = false },
  { { .name = "DO #1B" }, .state = false },
  { { .name = "DO #2A" }, .state = false },
  { { .name = "DO #2B" }, .state = false },
  { { .name = "DO #3A" }, .state = false },
  { { .name = "DO #3B" }, .state = false }
};

struct MENU_ENTRY menu_outputs[] =
{          //  ################# - 17 char limit for 20x4 LCD
  { { .text = "Digital Outputs:" }, .type = MENU_NOOP },
  { { .text = "DO #1A" }, .type = MENU_COMMAND, .data = (void *) gpio_output_menu, .user_data = 0, .extra = NULL },
  { { .text = "DO #1B" }, .type = MENU_COMMAND, .data = (void *) gpio_output_menu, .user_data = 1, .extra = NULL },
  { { .text = "DO #2A" }, .type = MENU_COMMAND, .data = (void *) gpio_output_menu, .user_data = 2, .extra = NULL },
  { { .text = "DO #2B" }, .type = MENU_COMMAND, .data = (void *) gpio_output_menu, .user_data = 3, .extra = NULL },
  { { .text = "DO #3A" }, .type = MENU_COMMAND, .data = (void *) gpio_output_menu, .user_data = 4, .extra = NULL },
  { { .text = "DO #3B" }, .type = MENU_COMMAND, .data = (void *) gpio_output_menu, .user_data = 5, .extra = NULL },
  { { .text = "" }, .type = MENU_NOTHING }
};

const PROGMEM struct MENU_TEXT_EXTRA extra_output_name =
{
  .text_size = 16
};

// Menu for editing saved networks
char gpio_edit_output_txt[8];
struct MENU_ENTRY menu_edit_output[] =
{
  { { .text = "" }, .type = MENU_LABEL, .data = (void *) &gpio_edit_output_txt, .user_data = 0, .extra = NULL },
  { { .text = "State: " }, .type = MENU_LABEL, .data = (void *) "N/A", .user_data = 0, .extra = NULL },
  { { .text = "Turn ON" }, .type = MENU_COMMAND, .data = (void *) gpio_set_output, .user_data = 1, .extra = NULL },
  { { .text = "Turn OFF" }, .type = MENU_COMMAND, .data = (void *) gpio_set_output, .user_data = 0, .extra = NULL },
  { { .text = "Edit name" }, .type = MENU_TEXT, .data = NULL, .user_data = 0, .extra = (void *) &extra_output_name },
  { { .text = "" }, .type = MENU_NOTHING }
};

extern unsigned long current_millis;
bool b_gpio_initialized = false;
unsigned long gpio_routine_millis = 2000;

// #####################################################################

void gpio_output_menu( struct MENU_ENTRY *menu_entry )
{
  // First line displays output name
  menu_edit_output[0].data = &gpio_digital_outputs[menu_entry->user_data].name;

  // Display state of the output
  if( gpio_digital_outputs[menu_entry->user_data].state == true )
    menu_edit_output[1].data = (void *) "ON";
  else
    menu_edit_output[1].data = (void *) "OFF";

  // Configure ON and OFF commands
  menu_edit_output[2].user_data = menu_entry->user_data;
  menu_edit_output[3].user_data = menu_entry->user_data + 10;

  // Store reference to output name for edit name command
  menu_edit_output[4].data = &gpio_digital_outputs[menu_entry->user_data].name;

  keyboard_push_stack( );

  // Set cursor to opposite menu entry to the current state
  if( gpio_digital_outputs[menu_entry->user_data].state == true )
  {
    // Set cursor point at "Turn OFF" menu item
    menu_cursor_position = 3;
  }
  else
  {
    // Set cursor point at "Turn ON" menu item
    menu_cursor_position = 2;
  }

  menu_screen_position = 0;
  menu_current = menu_edit_output;
  menu_current_size = 0;
}

// #####################################################################

void gpio_outputs_on_open( struct MENU_ENTRY *menu_entry )
{
  //
  // Update output names in the menu_gpio_outputs when user opens list of outputs
  //

  snprintf( menu_outputs[1].text, LCD_WIDTH - 3, "DO #1A: %s", gpio_digital_outputs[0].name );
  snprintf( menu_outputs[2].text, LCD_WIDTH - 3, "DO #1B: %s", gpio_digital_outputs[1].name );
  snprintf( menu_outputs[3].text, LCD_WIDTH - 3, "DO #2A: %s", gpio_digital_outputs[2].name );
  snprintf( menu_outputs[4].text, LCD_WIDTH - 3, "DO #2B: %s", gpio_digital_outputs[3].name );
  snprintf( menu_outputs[5].text, LCD_WIDTH - 3, "DO #3A: %s", gpio_digital_outputs[4].name );
  snprintf( menu_outputs[6].text, LCD_WIDTH - 3, "DO #3B: %s", gpio_digital_outputs[5].name );

}

// #####################################################################

void gpio_set_output( struct MENU_ENTRY *menu_entry )
{
  if( menu_entry->user_data < 10 )
    gpio_digital_outputs[menu_entry->user_data].state = true;
  else
    gpio_digital_outputs[menu_entry->user_data - 10].state = false;

  // Display state of the output and
  // change cursor position to opposite command
  if( gpio_digital_outputs[(menu_entry->user_data < 10 ? menu_entry->user_data : menu_entry->user_data - 10)].state == true )
  {
    menu_edit_output[1].data = (void *) "ON";
    menu_cursor_position = 3;
  }
  else
  {
    menu_edit_output[1].data = (void *) "OFF";
    menu_cursor_position = 2;
  }
}

// #####################################################################

void gpio_routine( )
{
  // check timer
  if( current_millis < gpio_routine_millis )
    return;
  gpio_routine_millis = current_millis + 1000;

  if( b_gpio_initialized == false )
  {
    // GPIO initialization
    Serial.println(F("[GPIO] Initializing PCF8575."));
    PCF.begin( 0xFFFF );

    bool status = PCF.isConnected( );
    if( status == true )
    {
      PCF.setButtonMask( 0b0111111 ); // 6 buttons
      
      Serial.println(F("[GPIO] PCF8575 is connected."));
      b_gpio_initialized = true;
    }
    else
    {
      Serial.println(F("[GPIO] PCF8575 is not connected."));
    }
  }
  else
  {
    // Check connection with GPIO chip
    bool status = PCF.isConnected( );
    if( status == false )
    {
      Serial.println(F("[GPIO] PCF8575 disconnected."));
      b_gpio_initialized = false;
    }

    // Update digital outputs
    if( gpio_digital_outputs[0].state == true ) { PCF.write( GPIO_OUTPUT_1A_PIN, LOW ); } else { PCF.write( GPIO_OUTPUT_1A_PIN, HIGH ); }
    if( gpio_digital_outputs[1].state == true ) { PCF.write( GPIO_OUTPUT_1B_PIN, LOW ); } else { PCF.write( GPIO_OUTPUT_1B_PIN, HIGH ); }
    if( gpio_digital_outputs[2].state == true ) { PCF.write( GPIO_OUTPUT_2A_PIN, LOW ); } else { PCF.write( GPIO_OUTPUT_2A_PIN, HIGH ); }
    if( gpio_digital_outputs[3].state == true ) { PCF.write( GPIO_OUTPUT_2B_PIN, LOW ); } else { PCF.write( GPIO_OUTPUT_2B_PIN, HIGH ); }
    if( gpio_digital_outputs[4].state == true ) { PCF.write( GPIO_OUTPUT_3A_PIN, LOW ); } else { PCF.write( GPIO_OUTPUT_3A_PIN, HIGH ); }
    if( gpio_digital_outputs[5].state == true ) { PCF.write( GPIO_OUTPUT_3B_PIN, LOW ); } else { PCF.write( GPIO_OUTPUT_3B_PIN, HIGH ); }

  }
}

// #####################################################################
