// #####################################################################

// Liquid Cristal Display

// include the library code:
#include <Arduino.h>

#include "lcd.h"
#include "menu.h"
#include "keyboard.h"

// initialize the library by associating any needed LCD interface pin
// with the arduino pin number it is connected to
const int rs = 16, en = 5, d4 = 18, d5 = 19, d6 = 20, d7 = 21;
LiquidCrystal lcd(rs, en, d4, d5, d6, d7);

#define LCD_RED     0
#define LCD_GREEN   1
#define LCD_BLUE    2

char display_state[LCD_WIDTH * LCD_HEIGHT];
char display_buffer[LCD_WIDTH * LCD_HEIGHT];

static const PROGMEM byte lcd_up_arrow_glyph[8] = {
  0b00100,
  0b01110,
  0b10101,
  0b00100,
  0b00100,
  0b00100,
  0b00000,
};

static const PROGMEM byte lcd_down_arrow_glyph[8] = {
  0b00000,
  0b00100,
  0b00100,
  0b00100,
  0b10101,
  0b01110,
  0b00100,
};

static const PROGMEM byte lcd_space_glyph[8] = {
  0b00000,
  0b00000,
  0b00000,
  0b00000,
  0b00000,
  0b10001,
  0b11111,
};

void lcd_change_color( uint8_t red, uint8_t green, uint8_t blue );
void lcd_change_color_command( struct MENU_ENTRY *menu_entry );
void lcd_change_color_on_change_event( struct MENU_ENTRY *menu_entry );

// Settings
int lcd_color[3];  // I could use uint8_t here, but currently only editing of "int" is implemented in the menu
int lcd_auto_powersave;
int lcd_auto_off;
bool lcd_always_on;
int lcd_animation;

bool lcd_off = false;

char lcd_refresh_cycles = 0;    // Prescale LCD refresh cycle to clear the display sometimes

//
// Animation
//

// Declare the colormap
struct colormap_t
{
    float val;
    float r;
    float g;
    float b;
};

// Jet colormap
/*static const PROGMEM struct colormap_t map_jet[] =
{
    { 0.00, 0.0, 0.0, 0.5 },
    { 0.20, 0.0, 1.0, 1.0 },
    { 0.40, 1.0, 1.0, 0.0 },
    { 0.60, 1.0, 0.0, 0.0 },
    { 0.80, 0.5, 0.0, 0.0 },
    { 1.00, 0.0, 0.0, 0.5 }
};*/

// Inverted viridis colormap
/*static const PROGMEM struct colormap_t map_viridis[] =
{
    { 0.00, 0.2667, 0.0039, 0.3294 },
    { 0.20, 0.2314, 0.3216, 0.5451 },
    { 0.40, 0.1294, 0.5686, 0.5490 },
    { 0.60, 0.3686, 0.7882, 0.3843 },
    { 0.80, 0.9922, 0.9059, 0.1451 },
    { 1.00, 0.2667, 0.0039, 0.3294 }
};*/

// All colors colormap
/*static const PROGMEM struct colormap_t map_all_colors[] =
{
    { 0.000000000000000, 0.0, 0.0, 1.0 },
    { 0.142857142857143, 0.0, 1.0, 0.0 },
    { 0.285714285714286, 0.0, 1.0, 1.0 },
    { 0.428571428571429, 1.0, 0.0, 0.0 },
    { 0.571428571428572, 1.0, 0.0, 1.0 },
    { 0.714285714285715, 1.0, 1.0, 0.0 },
    { 0.857142857142857, 1.0, 1.0, 1.0 },
    { 1.000000000000000, 0.0, 0.0, 1.0 }
};*/

// Rainbow colormap
static const PROGMEM struct colormap_t map_rainbow[] =
{
    { 0.000000000000000, 0.578125, 0.0, 0.82421875 },
    { 0.142857142857143, 0.29296875, 0.0, 0.5078125 },
    { 0.285714285714286, 0.0, 0.0, 1.0 },
    { 0.428571428571429, 0.0, 1.0, 0.0 },
    { 0.571428571428572, 1.0, 1.0, 0.0 },
    { 0.714285714285715, 1.0, 0.5, 0.0 },
    { 0.857142857142857, 1.0, 0.0, 0.0 },
    { 1.000000000000000, 0.578125, 0.0, 0.82421875 }
};

static const PROGMEM struct LCD_ANIMATION
{
  const struct colormap_t *map_ptr;
  size_t map_size;
}
lcd_animation_list[] =
{
  { map_rainbow, sizeof(map_rainbow) }
  //{ map_viridis, sizeof(map_viridis) },
  //{ map_jet, sizeof(map_jet) },
  //{ map_all_colors, sizeof(map_all_colors) }
};

void colormap( float val, float min, float max, unsigned char *r, unsigned char *g, unsigned char *b, struct colormap_t *map, size_t map_size );

const PROGMEM struct MENU_INTEGER_EXTRA extra_color =
{
  .on_change = lcd_change_color_on_change_event,
  { .units = "" },
  .minimum = 0,
  .maximum = 255,
  .increment = 1,
  .increment_fast = 5
};

const PROGMEM struct MENU_ENTRY menu_lcd_color[] =
{          //  ################# - 17 char limit for 20x4 LCD
  { { .text = "White" }, .type = MENU_COMMAND, .data = (void *) &lcd_change_color_command, .user_data = 0b111, .extra = NULL },
  { { .text = "Off" }, .type = MENU_COMMAND, .data = (void *) &lcd_change_color_command, .user_data = 0b000, .extra = NULL },
  { { .text = "Red" }, .type = MENU_COMMAND, .data = (void *) &lcd_change_color_command, .user_data = 0b100, .extra = NULL },
  { { .text = "Green" }, .type = MENU_COMMAND, .data = (void *) &lcd_change_color_command, .user_data = 0b010, .extra = NULL },
  { { .text = "Blue" }, .type = MENU_COMMAND, .data = (void *) &lcd_change_color_command,  .user_data = 0b001, .extra = NULL },
  { { .text = "Yellow" }, .type = MENU_COMMAND, .data = (void *) &lcd_change_color_command,  .user_data = 0b110, .extra = NULL },
  { { .text = "Magenta" }, .type = MENU_COMMAND, .data = (void *) &lcd_change_color_command, .user_data = 0b101, .extra = NULL },
  { { .text = "Cyan" }, .type = MENU_COMMAND, .data = (void *) &lcd_change_color_command, .user_data = 0b011, NULL },
  { { .text = "Custom Red" }, .type = MENU_INTEGER, .data = (void *) &lcd_color[0], .user_data = 0, .extra = (void *) &extra_color },
  { { .text = "Custom Green" }, .type = MENU_INTEGER, .data = (void *) &lcd_color[1], .user_data = 1, .extra = (void *) &extra_color },
  { { .text = "Custom Blue" }, .type = MENU_INTEGER, .data = (void *) &lcd_color[2], .user_data = 2, .extra = (void *) &extra_color },
  { { .text = "" }, .type = MENU_NOTHING }
};

const PROGMEM struct MENU_INTEGER_EXTRA extra_timeout2 =
{
  .on_change = NULL,
  { .units = "s" },
  .minimum = 0,
  .maximum = 3600,
  .increment = 1,
  .increment_fast = 5
};

const PROGMEM struct MENU_INTEGER_EXTRA extra_animation =
{
  .on_change = NULL,
  { .units = "" },
  .minimum = 0,
  .maximum = sizeof(lcd_animation_list) / sizeof(struct LCD_ANIMATION),
  .increment = 1,
  .increment_fast = 1
};

const PROGMEM struct MENU_ENTRY menu_lcd[] =
{          //  ################# - 17 char limit for 20x4 LCD
  { { .text = "[Color]" }, .type = MENU_SUBMENU, .data = (void *) &menu_lcd_color, .user_data = 0, .extra = NULL },
  { { .text = "Animation" }, .type = MENU_INTEGER, .data = (void *) &lcd_animation, .user_data = 0, .extra = (void *) &extra_animation },
  { { .text = "Auto-Powersave" }, .type = MENU_INTEGER, .data = (void *) &lcd_auto_powersave, .user_data = 0, .extra = (void *) &extra_timeout2 },
  { { .text = "Auto-Off" }, .type = MENU_INTEGER, .data = (void *) &lcd_auto_off, .user_data = 0, .extra = (void *) &extra_timeout2 },
  { { .text = "Always-On" }, .type = MENU_BOOL, .data = &lcd_always_on, .user_data = 0, NULL },
  { { .text = "" }, .type = MENU_NOTHING }
};

extern unsigned long current_millis;
bool b_lcd_initialized = false;
unsigned long lcd_routine_millis = 500;

// #####################################################################

void lcd_change_color( uint8_t red, uint8_t green, uint8_t blue )
{
  analogWrite( LCD_RED, red );
  analogWrite( LCD_GREEN, green );
  analogWrite( LCD_BLUE, blue );
}

// #####################################################################

void lcd_change_color_command( struct MENU_ENTRY *menu_entry )
{
  uint8_t color = menu_entry->user_data;

  lcd_color[0] = color & 0b100 ? 255 : 0;
  lcd_color[1] = color & 0b010 ? 255 : 0;
  lcd_color[2] = color & 0b001 ? 255 : 0;
  
  lcd_change_color(
    lcd_color[0],
    lcd_color[1],
    lcd_color[2]
  );
}

// #####################################################################

void lcd_change_color_on_change_event( struct MENU_ENTRY *menu_entry )
{
  lcd_color[menu_entry->user_data] = menu_original_integer;
}

// #####################################################################

void colormap( float val, float min, float max, unsigned char *r, unsigned char *g, unsigned char *b, const struct colormap_t *map, size_t map_size )
{
    const int map_n = map_size / sizeof(struct colormap_t);

    // Normalise the value
    double val_norm = (val - min) / (max - min);

    // Find interal in the colormap
    int i;
    for( i = 0; i < map_n - 1; i++ )
    {
        if( map[i].val <= val_norm &&
            map[i + 1].val >= val_norm )
        {
            // Found
            goto colormap_found;
        }
    }

    // Not found
    if( map[0].val > val_norm )
    {
        *r = map[0].r * 255.0;
        *g = map[0].g * 255.0;
        *b = map[0].b * 255.0;
    }
    else if( map[map_n - 1].val < val_norm )
    {
        *r = map[map_n - 1].r * 255.0;
        *g = map[map_n - 1].g * 255.0;
        *b = map[map_n - 1].b * 255.0;
    }

    return;

    float val_pos;

colormap_found:

    // Calculate the relative position of the value in the interval
    val_pos = (val_norm - map[i].val) / (map[i + 1].val - map[i].val);

    // Interpolate color component according to val_pos
    float rd = val_pos * (map[i + 1].r - map[i].r) + map[i].r;
    float gd = val_pos * (map[i + 1].g - map[i].g) + map[i].g;
    float bd = val_pos * (map[i + 1].b - map[i].b) + map[i].b;

    // Output
    *r = rd * 255.0;
    *g = gd * 255.0;
    *b = bd * 255.0;

    //printf( "val=%g, val_norm=%g, val_pos=%g, i=%i, rgb = %g %g %g\n", val, val_norm, val_pos, i, rd, gd, bd );
}

// #####################################################################

void lcd_routine( )
{
  // check timer
  if( current_millis < lcd_routine_millis )
    return;
  lcd_routine_millis = current_millis + 250;

  if( b_lcd_initialized == false )
  {
    // LCD initialization
    Serial.println(F("[LCD] Initializing Winstar Display WH2004A-CFH-JT#."));

    // Store custom characters
    lcd.createChar( 0, (uint8_t *) lcd_up_arrow_glyph );
    lcd.createChar( 1, (uint8_t *) lcd_down_arrow_glyph );
    lcd.createChar( 2, (uint8_t *) lcd_space_glyph );

    // set up the LCD's number of columns and rows:
    lcd.begin(LCD_WIDTH, LCD_HEIGHT);
    lcd.setCursor(0, 0);

    // Backlight initialization
    pinMode( LCD_RED, OUTPUT );
    pinMode( LCD_GREEN, OUTPUT );
    pinMode( LCD_BLUE, OUTPUT );
    lcd_change_color(
      lcd_color[0],
      lcd_color[1],
      lcd_color[2]
    );

    // Test
    lcd.setCursor(0, 0);
    lcd.print("-(__^> PIK-PIK RTU2");
    memset( display_buffer, ' ', sizeof(display_buffer) );
    memcpy( display_buffer, "-(__^> PIK-PIK RTU2", 19 );
    //lcd.setCursor(0, 2);
    //lcd.print("Test!");

    b_lcd_initialized = true;
  }
  else
  {
    // Reset display before rendering
    //lcd.clear( );
    //lcd.noCursor( );

    // Turn off backlight after "lcd_auto_powersave" seconds
    if( lcd_always_on == false && key_last_activity + (1000 * lcd_auto_powersave) < current_millis )
      lcd_change_color( 0, 0, 0 );  // TODO: maybe make a smooth fade out animation?  
    else
    {
      if( lcd_animation > 0 )
      {
        unsigned char r, g, b;

        // This function is ported from audiovizsdl
        // It is very ineffective... but it works!
        const struct colormap_t *map_ptr = lcd_animation_list[lcd_animation - 1].map_ptr;
        size_t map_size = lcd_animation_list[lcd_animation - 1].map_size;
        size_t map_num = map_size / sizeof(struct colormap_t);
        colormap(
          current_millis % (5 * map_num * 1000), 0, 5 * map_num * 1000.0,
          &r, &g, &b,
          map_ptr,
          map_size
        );
        
        lcd_change_color( r, g, b );
      }
      else
      {
        lcd_change_color( lcd_color[0], lcd_color[1], lcd_color[2] );
      }
    }

    // Turn off LCD completely and stop updating it
    if( key_last_activity + (1000 * lcd_auto_off) < current_millis && lcd_always_on == false )
    {
      if( lcd_off == false )
      {
        lcd.noDisplay( );
        lcd_off = true;
      }
      return;
    }
    if( lcd_off == true )
    {
      lcd.display( );
      lcd_off = false;
    }

    //
    // Dirty hack/workaround
    //
    // Electromagnetic interefence messes up communication between Arduino and
    // chinese LCD makeing it display garbage.
    //
    // Periodic clearing of display brings everything back to normal.
    //
    if( lcd_refresh_cycles >= 10 )
    {
      lcd_refresh_cycles = 0;
      lcd.clear( );
      lcd.noCursor( );
    }
    else
    {
      lcd_refresh_cycles++;
    }

    // Update the display
    for( int y = 0; y < LCD_HEIGHT; y++ )
    {
      int prev_x = 0;
      for( int x = 0; x < LCD_WIDTH; x++ )
      {
        char ch = display_buffer[x + y * LCD_WIDTH];
  
        // Try to update differences only
        //if( display_state[x + y * LCD_WIDTH] != ch )  // TODO: doesn't work very well :(
        {
          // If this a concsecutive stream of characters,
          // then no need to update cursor position again
          if( prev_x + 1 != x )
          {
            lcd.setCursor( x, y );
          }

          display_state[x + y * LCD_WIDTH] = ch;
          lcd.write( ch );

          prev_x = x;
        }
      }
    }
  }
}

// #####################################################################
