// #####################################################################

// Menu subsystem

#include "menu.h"
#include "lcd.h"
#include "led.h"
#include "wifi.h"
#include "beep.h"
#include "ntp.h"
#include "datalogger.h"
#include "ds18b20.h"
#include "gpio.h"
#include "sd.h"
#include "pmic.h"
#include "watchdog.h"
#include "sensor.h"
#include "main.h"
#include "keyboard.h"

//extern "C"
//{

//
// Before menu optimization
// Скетч использует 158996 байт (60%) памяти устройства. Всего доступно 262144 байт.
// Глобальные переменные используют 27232 байт (83%) динамической памяти, оставляя 5536 байт для локальных переменных. Максимум: 32768 байт.
//
// After menu optimization
// Скетч использует 150596 байт (57%) памяти устройства. Всего доступно 262144 байт.
// Глобальные переменные используют 21568 байт (65%) динамической памяти, оставляя 11200 байт для локальных переменных. Максимум: 32768 байт.
//

const PROGMEM struct MENU_ENTRY menu_info[] =
{
  { { .text = "Version: 0.8" }, .type = MENU_NOOP },
  { { .text = "Compiled: " }, .type = MENU_NOOP },
  { { .text = __DATE__ }, .type = MENU_NOOP },
  { { .text = __TIME__ }, .type = MENU_NOOP },
  { { .text = "" }, .type = MENU_NOTHING }
};

const PROGMEM struct MENU_SUBMENU_EXTRA extra_outputs =
{
  .on_open = &gpio_outputs_on_open
};

void test( struct MENU_ENTRY *menu_entry );
const PROGMEM struct MENU_ENTRY menu_main[] =
{          //  ################# - 17 char limit for 20x4 LCD
  { { .text = "Test" }, .type = MENU_COMMAND, .data = (void *) test, .user_data = 0, .extra = NULL },
  { { .text = "[LCD]" }, .type = MENU_SUBMENU, .data = (void *) &menu_lcd, .user_data = 0, .extra = NULL },
  { { .text = "[LEDs]" }, .type = MENU_SUBMENU, .data = (void *) &menu_led, .user_data = 0, .extra = NULL },
  { { .text = "[WiFi]" }, .type = MENU_SUBMENU, .data = (void *) &menu_wifi, .user_data = 0, .extra = NULL },
  { { .text = "[Beep]" }, .type = MENU_SUBMENU, .data = (void *) &menu_beep, .user_data = 0, .extra = NULL },
  { { .text = "[RTC]" }, .type = MENU_SUBMENU, .data = (void *) &menu_rtc, .user_data = 0, .extra = NULL },
  { { .text = "[Datalogger]" }, .type = MENU_SUBMENU, .data = (void *) &menu_datalogger, .user_data = 0, .extra = NULL },
  { { .text = "[Webserver]" }, .type = MENU_NOOP },
  { { .text = "[Inputs]" }, .type = MENU_SUBMENU, .data = (void *) &menu_inputs, .user_data = 0, .extra = NULL },
  { { .text = "[Outputs]" }, .type = MENU_SUBMENU, .data = (void *) &menu_outputs, .user_data = 0, .extra = (void *) &extra_outputs },
  { { .text = "[SD Card]" }, .type = MENU_SUBMENU, .data = (void *) &menu_sd, .user_data = 0, .extra = NULL },
  { { .text = "[PMIC]" }, .type = MENU_SUBMENU, .data = (void *) &menu_battery, .user_data = 0, .extra = NULL },
  { { .text = "[Info]" }, .type = MENU_SUBMENU, .data = (void *) &menu_info, .user_data = 0, .extra = NULL },
  { { .text = "[Watchdog]" }, .type = MENU_SUBMENU, .data = (void *) &menu_watchdog, .user_data = 0, .extra = NULL },
  { { .text = "[ENV Shield]" }, .type = MENU_SUBMENU, .data = (void *) &menu_env_shield, .user_data = 0, .extra = NULL },
  { { .text = "Store Settings" }, .type = MENU_COMMAND, .data = (void *) sd_store_configuration_command, .user_data = 0, .extra = NULL },
  { { .text = "Read Settings" }, .type = MENU_COMMAND, .data = (void *) sd_read_configuration_command, .user_data = 0, .extra = NULL },
  { { .text = "Factory default" }, .type = MENU_COMMAND, .data = (void *) factory_default, .user_data = 0, .extra = NULL },
  { { .text = "Restart" }, .type = MENU_COMMAND, .data = (void *) restart, .user_data = 0, .extra = NULL },
  { { .text = "Sleep" }, .type = MENU_NOOP },
  { { .text = "" }, .type = MENU_NOTHING }
};

int menu_original_integer = 0;
int menu_integer_position = 0;
struct MENU_ENTRY menu_edit_integer[] =
{
  { { .text = "" }, .type = MENU_LABEL, .data = NULL, .user_data = 0, .extra = NULL },
  { { .text = "" }, .type = MENU_NOOP },
  { { .text = "" }, .type = MENU_EDIT_INTEGER },
  { { .text = "" }, .type = MENU_NOTHING }
};

bool menu_original_bool = false;
struct MENU_ENTRY menu_edit_bool[] =
{
  { { .text = "" }, .type = MENU_LABEL, .data = NULL, .user_data = 0, .extra = NULL },
  { { .text = "" }, .type = MENU_NOOP },
  { { .text = "" }, .type = MENU_EDIT_BOOL },
  { { .text = "" }, .type = MENU_NOTHING }
};

float menu_original_float = 0.0;
int menu_float_position = 0;
struct MENU_ENTRY menu_edit_float[] =
{
  { { .text = "" }, .type = MENU_LABEL, .data = NULL, .user_data = 0, .extra = NULL },
  { { .text = "" }, .type = MENU_NOOP },
  { { .text = "" }, .type = MENU_EDIT_FLOAT },
  { { .text = "" }, .type = MENU_NOTHING }
};

char menu_original_text[256];
int menu_text_position = 0;
int menu_text_scroll = 0;   // For long text that doesn't fit in LCD_WIDTH
struct MENU_ENTRY menu_edit_text[] =
{
  { { .text = "" }, .type = MENU_LABEL, .data = NULL, .user_data = 0, .extra = NULL },
  { { .text = "" }, .type = MENU_NOOP },
  { { .text = "" }, .type = MENU_EDIT_TEXT },
  { { .text = "" }, .type = MENU_NOTHING }
};

void menu_display_messagebox( );
void menu_messagebox_command( struct MENU_ENTRY *menu_entry );
struct MENU_ENTRY menu_messagebox[] =
{
  { { .text = "" }, .type = MENU_NOOP },
  { { .text = "" }, .type = MENU_NOOP },
  { { .text = "" }, .type = MENU_NOOP },
  { { .text = "[OK]" }, .type = MENU_COMMAND, .data = (void *) menu_messagebox_command, .user_data = 1, .extra = NULL },
  { { .text = "" }, .type = MENU_NOTHING }
};

struct MENU_ENTRY *menu_current = NULL;
int menu_current_size = 0;

int menu_horizontal_scroll = 0; // For scrolling long texts that don't fit in LCD_WIDTH
int menu_cursor_position = 0;   // Tells which menu item is selected
int menu_screen_position = 0;   // Tells which part of the menu to render on the screen

enum MENU_MODE menu_mode = MENU_NORMAL;

// Menu stack for "ESC" button to work as "Back" function,
// when deep in sub-menus
struct MENU_STACK menu_stack[10];
int menu_stack_position = 0;

extern unsigned long current_millis;
bool b_menu_initialized = false;
unsigned long menu_routine_millis = 250;

//}

// #####################################################################

void menu_display_messagebox( )
{
  // Open messagebox
  keyboard_push_stack( );
  menu_cursor_position = 3;
  menu_screen_position = 0;
  menu_current = menu_messagebox;
  menu_current_size = 0;
}

// #####################################################################

void menu_messagebox_command( struct MENU_ENTRY *menu_entry )
{
  // Simulate "ESC" key press
  keyboard_event( KEY_ESC, KEYFLAG_PRESS );
}

// #####################################################################

void menu_routine( )
{
  // check timer
  if( current_millis < menu_routine_millis )
    return;
  menu_routine_millis = current_millis + 250;

  if( b_menu_initialized == false )
  {
    if( b_gpio_initialized == true && b_lcd_initialized == true )
    {
      memset( &menu_stack, 0, sizeof(menu_stack) );
      
      Serial.println(F("[Menu] Menu subsystem initialized."));
      b_menu_initialized = true;
    }
  }
  else
  {
    // Clear the LCD buffer first
    memset( display_buffer, ' ', sizeof(display_buffer) );

    if( menu_current == NULL )
    {
      //
      // Main screen
      //

      // Print Date/Time
      time_t epoch = rtc.getEpoch( ) + rtc_timezone * 60 * 60; // timezone correction
      struct tm* tmp = gmtime( &epoch ); // struct tm is a structure that contains a calendar time broken down into its components
      char txt_buffer[32];
      size_t txt_size = strftime( txt_buffer, sizeof(txt_buffer), "%d.%m.%Y %H:%M:%S", tmp );
      if( txt_size > LCD_WIDTH ) txt_size = LCD_WIDTH;
      memcpy( &display_buffer[0 + 0 * LCD_WIDTH], txt_buffer, txt_size );
    
      //int txt_size2 = snprintf( txt_buffer, sizeof(txt_buffer), "%+5.1f \xDF%nC %5.2f %% rH", env_temperature, env_humidity );
      int txt_size2 = snprintf( txt_buffer, sizeof(txt_buffer), "%+5.1f " "\xDF" "C %5.2f %% rH", env_temperature, env_humidity );
      //int txt_size2 = snprintf( txt_buffer, sizeof(txt_buffer), "%5.1f %5.1f", env_temperature, env_humidity );
      if( txt_size2 > LCD_WIDTH ) txt_size2 = LCD_WIDTH;
      memcpy( &display_buffer[0 + 1 * LCD_WIDTH], txt_buffer, txt_size2 );
    
      txt_size2 = snprintf( txt_buffer, sizeof(txt_buffer), "%5.2f kPa %4.2f V", env_pressure, battery_voltage );
      if( txt_size2 > LCD_WIDTH ) txt_size2 = LCD_WIDTH;
      memcpy( &display_buffer[0 + 2 * LCD_WIDTH], txt_buffer, txt_size2 );
      
      memcpy( &display_buffer[0 + 3 * LCD_WIDTH], "-(__^> PIK-PIK RTU2", 19 );
    }
    else
    {
      //
      // Menu renderer
      //
      
      // Count menu items
      // Calculation happen only once, when menu is changed
      if( menu_current_size == 0 )
      for( int i = 0; menu_current[i].type != MENU_NOTHING; i++ )
        menu_current_size++;

      // Test
      //char txt_buffer[32];
      //int txt_size = snprintf( txt_buffer, sizeof(txt_buffer), "%i", menu_current_size );
      //memcpy( &display_buffer[10 + 1 * LCD_WIDTH], txt_buffer, txt_size );

      // Render current menu
      for( int i = 0; menu_current[i].type != MENU_NOTHING; i++ )
      {
        // Skip menu items that are above the screen
        if( menu_screen_position > i )
          continue;

        // Limit printing only LCD_HEIGHT number of menu items
        if( i - menu_screen_position >= LCD_HEIGHT )
          break;

        if( menu_current[i].type == MENU_EDIT_INTEGER )
        {
          int value;
          if( menu_mode == MENU_EDITING_INTEGER )
            value = menu_original_integer;
          else
            value = *((int *) ((struct MENU_ENTRY *) menu_current[i].data)->data);
          
          char *units = &( (struct MENU_INTEGER_EXTRA *) ((struct MENU_ENTRY *) menu_current[i].data)->extra )->units[0];
          
          char txt_buffer[LCD_WIDTH + 1];
          int txt_size = snprintf( txt_buffer, sizeof(txt_buffer), "%i", value );
          int txt_size2 = snprintf( txt_buffer, sizeof(txt_buffer), "%i %s", value, units );
          //                          \/ this formula shifts text of integer to the right, when position is more than text length
          memcpy( &display_buffer[2 + (menu_integer_position + 1 > txt_size ? menu_integer_position - txt_size + 1 : 0) + (i - menu_screen_position) * LCD_WIDTH], txt_buffer, txt_size2 );
        }
        else if( menu_current[i].type == MENU_EDIT_BOOL )
        {
          bool value;
          if( menu_mode == MENU_EDITING_BOOL )
            value = menu_original_bool;
          else
            value = *((bool *) ((struct MENU_ENTRY *) menu_current[i].data)->data);
            
          if( value == true )
            memcpy( &display_buffer[2 + (i - menu_screen_position) * LCD_WIDTH], "TRUE", 4 );
          else if( value == false )
            memcpy( &display_buffer[2 + (i - menu_screen_position) * LCD_WIDTH], "FALSE", 5 );
        }
        else if( menu_current[i].type == MENU_EDIT_FLOAT )
        {
          struct MENU_ENTRY *menu_entry = (struct MENU_ENTRY *) menu_current[i].data;
          float value;
          if( menu_mode == MENU_EDITING_FLOAT )
            value = menu_original_float;
          else
            value = *((float *) menu_entry->data);
          char *units = &((struct MENU_FLOAT_EXTRA *) menu_entry->extra)->units[0];

          // Convert float into readable text
          char txt_buffer[LCD_WIDTH + 1];
          char txt_buffer2[LCD_WIDTH + 1];
          snprintf( txt_buffer, sizeof(txt_buffer), "%%0.%if", ((struct MENU_FLOAT_EXTRA *) menu_entry->extra)->precision );
          int txt_size3 = snprintf( txt_buffer2, sizeof(txt_buffer2), txt_buffer, value ); //   <- this is just to calculate lenght of text without units
          snprintf( txt_buffer, sizeof(txt_buffer), "%%0.%if %s", ((struct MENU_FLOAT_EXTRA *) menu_entry->extra)->precision, units );
          int txt_size2 = snprintf( txt_buffer2, sizeof(txt_buffer2), txt_buffer, value );

          // Find location of the dot '.'
          int dot_location = 0;
          for( int i = 0; txt_buffer2[i] != 0; i++ )
          {
            if( txt_buffer2[i] == '.' )
            {
              dot_location = i;
              break;
            }
          }

          if( dot_location == 0 )
            dot_location = txt_size3;
          
          memcpy( &display_buffer[2 + (menu_float_position + 1 > dot_location ? menu_float_position - dot_location + 1 : 0) + (i - menu_screen_position) * LCD_WIDTH], txt_buffer2, txt_size2 );
        }
        else if( menu_current[i].type == MENU_EDIT_TEXT )
        {
          struct MENU_ENTRY *menu_entry = (struct MENU_ENTRY *) menu_current[i].data;
          char *text_buf = (char *) menu_entry->data;
          if( menu_mode == MENU_EDITING_TEXT )
            text_buf = menu_original_text;
          else
            text_buf = (char *) menu_entry->data;
          int text_len = strlen( text_buf );

          // Limit displayed characters
          if( text_len - menu_text_scroll > LCD_WIDTH - 3 )
            text_len = LCD_WIDTH - 3;
          else
            text_len -= menu_text_scroll;

          if( text_len > 0 )
          {
            if( menu_mode == MENU_EDITING_TEXT )
            {
              // Render the text, but display special space character glyph instead of ordinary spaces
              for( int j = 0; j < text_len; j++ )
              {
                char ch = *(text_buf + menu_text_scroll + j);
                if( ch == 0x020 ) ch = 0x002;
                display_buffer[2 + (i - menu_screen_position) * LCD_WIDTH + j] = ch;
              }
            }
            else
            {
              // Normal text render
              memcpy( &display_buffer[2 + (i - menu_screen_position) * LCD_WIDTH], text_buf + menu_text_scroll, text_len );
            }
          }

          // Also display current cursor posistion and maximum text length
          char position_txt[LCD_WIDTH + 1];
          int position_size = snprintf( position_txt, sizeof(position_txt), "%i/%i", menu_text_position + 1, ((struct MENU_TEXT_EXTRA *) menu_entry->extra)->text_size );
          memcpy( &display_buffer[2 + (i - menu_screen_position - 1) * LCD_WIDTH], position_txt, position_size );
        }
        else
        {
          char large_buffer[256];

          // Main text
          int txt_size = strlen( menu_current[i].text );
          strncpy( large_buffer, &menu_current[i].text[0], sizeof(large_buffer) );

          // Add extra label text
          if( txt_size < sizeof(large_buffer) && menu_current[i].type == MENU_LABEL && menu_current[i].data != NULL )
          {
            strncat( large_buffer, (char *) menu_current[i].data, sizeof(large_buffer) );
          }

          txt_size = strlen( large_buffer );

          if( i == menu_cursor_position )
          {
            // Limit displayed characters
            if( txt_size - menu_horizontal_scroll > LCD_WIDTH - 3 )
              txt_size = LCD_WIDTH - 3;
            else
              txt_size -= menu_horizontal_scroll;
  
            if( txt_size > 0 )
              memcpy( &display_buffer[2 + (i - menu_screen_position) * LCD_WIDTH], large_buffer + menu_horizontal_scroll, txt_size );
          }
          else
          {
            // Limit displayed characters
            if( txt_size > LCD_WIDTH - 3 )
              txt_size = LCD_WIDTH - 3;
  
            if( txt_size > 0 )
              memcpy( &display_buffer[2 + (i - menu_screen_position) * LCD_WIDTH], large_buffer, txt_size );
          }

          // Old renderer, without scroll
          // Draw menu item text in the display buffer
          /*size_t txt_size = strlen( menu_current[i].text );
          memcpy( &display_buffer[2 + (i - menu_screen_position) * LCD_WIDTH], &menu_current[i].text, txt_size );

          if( menu_current[i].type == MENU_LABEL && menu_current[i].data != NULL )
          {
            // Also print LABEL data
            size_t txt_size2 = strlen( (char *) menu_current[i].data );
            memcpy( &display_buffer[2 + txt_size + (i - menu_screen_position) * LCD_WIDTH], (char *) menu_current[i].data, txt_size2 );
          }*/
        }
      }

      // Render menu cursor
      if( menu_mode == MENU_NORMAL )
      {
        display_buffer[0 + (menu_cursor_position - menu_screen_position) * LCD_WIDTH] = 0x07E; // 0x07E = "->" character
      }
      else if( menu_mode == MENU_EDITING_INTEGER )
      {
        // Display other selection character, which indicates that the value is being edited
        display_buffer[0 + (menu_cursor_position - menu_screen_position) * LCD_WIDTH] = 0x0FF; // 0x0FF = "█" character

        // Display position character "^" below the integer that is being edited
        struct MENU_ENTRY *menu_entry = (struct MENU_ENTRY *) menu_current[menu_cursor_position].data;
        char txt_buffer[LCD_WIDTH + 1];
        int txt_size = snprintf( txt_buffer, sizeof(txt_buffer), "%i", menu_original_integer /* *((int *) menu_entry[0].data) */ );
        if( menu_integer_position >= txt_size )
          display_buffer[2 + (menu_cursor_position - menu_screen_position + 1) * LCD_WIDTH] = '^';
        else
          display_buffer[txt_size + 1 - menu_integer_position + (menu_cursor_position - menu_screen_position + 1) * LCD_WIDTH] = '^';
      }
      else if( menu_mode == MENU_EDITING_BOOL )
      {
        // Display other selection character, which indicates that the value is being edited
        display_buffer[0 + (menu_cursor_position - menu_screen_position) * LCD_WIDTH] = 0x0FF; // 0x0FF = "█" character

        bool value = menu_original_bool; //*((bool *) ((struct MENU_ENTRY *) menu_current[menu_cursor_position].data)->data);
        if( value == false )
          display_buffer[2 + 4 + (menu_cursor_position - menu_screen_position + 1) * LCD_WIDTH] = '^';
        else
          display_buffer[2 + 3 + (menu_cursor_position - menu_screen_position + 1) * LCD_WIDTH] = '^';
      }
      else if( menu_mode == MENU_EDITING_FLOAT )
      {
        // Display other selection character, which indicates that the value is being edited
        display_buffer[0 + (menu_cursor_position - menu_screen_position) * LCD_WIDTH] = 0x0FF; // 0x0FF = "█" character

        // Display position character "^" below the float that is being edited
        struct MENU_ENTRY *menu_entry = (struct MENU_ENTRY *) menu_current[menu_cursor_position].data;

        // Convert float to text
        char txt_buffer[LCD_WIDTH + 1];
        char txt_buffer2[LCD_WIDTH + 1];
        snprintf( txt_buffer, sizeof(txt_buffer), "%%0.%if", ((struct MENU_FLOAT_EXTRA *) menu_entry->extra)->precision );
        int txt_size2 = snprintf( txt_buffer2, sizeof(txt_buffer2), txt_buffer, menu_original_float /* *((float *) menu_entry[0].data) */ );

        // Find location of the dot '.'
        int dot_location = 0;
        while( 1 )
        {
          if( txt_buffer2[dot_location] == 0 || txt_buffer2[dot_location] == '.' )
            break;

          dot_location++;
        }

        if( menu_float_position >= dot_location )
          display_buffer[2 + (menu_cursor_position - menu_screen_position + 1) * LCD_WIDTH] = '^';
        else if( menu_float_position < 0 )
          display_buffer[dot_location + 1 - menu_float_position + 1 + (menu_cursor_position - menu_screen_position + 1) * LCD_WIDTH] = '^';
        else
          display_buffer[dot_location + 1 - menu_float_position + (menu_cursor_position - menu_screen_position + 1) * LCD_WIDTH] = '^';
      }
      else if( menu_mode == MENU_EDITING_TEXT )
      {
        // Display other selection character, which indicates that the value is being edited
        display_buffer[0 + (menu_cursor_position - menu_screen_position) * LCD_WIDTH] = 0x0FF; // 0x0FF = "█" character

        // Display position character "^" below the text that is being edited
        struct MENU_ENTRY *menu_entry = (struct MENU_ENTRY *) menu_current[menu_cursor_position].data;
        if( menu_text_position - menu_text_scroll < LCD_WIDTH - 3 )
          display_buffer[2 + menu_text_position - menu_text_scroll + (menu_cursor_position - menu_screen_position + 1) * LCD_WIDTH] = '^';
        else
          display_buffer[2 + LCD_WIDTH - 3 - 1 + (menu_cursor_position - menu_screen_position + 1) * LCD_WIDTH] = '^';
      }

      // Render up and down arrows showing that there are more
      // menu item above or below
      if( menu_screen_position > 0 )
        display_buffer[LCD_WIDTH - 1 + 0 * LCD_WIDTH] = 0x000;
      if( menu_current_size - menu_screen_position - LCD_HEIGHT > 0 )
        display_buffer[LCD_WIDTH - 1 + (LCD_HEIGHT - 1) * LCD_WIDTH] = 0x001;

      // Show current position and number of menu entries in the corner
      char txt_buffer[32];
      int txt_size = snprintf( txt_buffer, sizeof(txt_buffer), "%i/%i", menu_cursor_position + 1, menu_current_size );
      memcpy( &display_buffer[LCD_WIDTH - 1 - txt_size + 0 * LCD_WIDTH], txt_buffer, txt_size );
    }
  }
}

// #####################################################################
