// #####################################################################

// Menu subsystem

#include "lcd.h"
#include "gpio.h"

enum MENU_TYPE
{
  MENU_NOTHING = 0,   // null termination menu type
  MENU_NOOP,
  MENU_LABEL,
  MENU_COMMAND,
  MENU_SUBMENU,
  MENU_INTEGER,
  MENU_EDIT_INTEGER,
  MENU_BOOL,
  MENU_EDIT_BOOL,
  MENU_FLOAT,
  MENU_EDIT_FLOAT,
  MENU_TEXT,
  MENU_EDIT_TEXT
};

// #####################################################################

struct MENU_ENTRY
{
  char text[LCD_WIDTH + 1];
  enum MENU_TYPE type;
  void *data;
  int user_data;

  void *extra;
};

// #####################################################################

struct MENU_SUBMENU_EXTRA
{
  // Event handlers
  void (*on_open)(struct MENU_ENTRY *);
};

struct MENU_INTEGER_EXTRA
{
  void (*on_change)(struct MENU_ENTRY *);

  // For MENU_INTEGER
  char units[8];
  int minimum;
  int maximum;
  int increment;
  int increment_fast;
};

struct MENU_FLOAT_EXTRA
{
  void (*on_change)(struct MENU_ENTRY *);

  // For MENU_FLOAT
  char units[8];
  int precision;
  float f_minimum;
  float f_maximum;
  float f_increment;
  float f_increment_fast;
};

struct MENU_TEXT_EXTRA
{
  // For MENU_TEXT
  int text_size;
};

// #####################################################################

// Menu stack for "ESC" button to work as "Back" function,
// when deep in sub-menus
struct MENU_STACK
{
  struct MENU_ENTRY *entry;
  int cursor_position;
  int screen_position;
};

extern struct MENU_STACK menu_stack[10];

extern int menu_stack_position;

// #####################################################################

extern struct MENU_ENTRY *menu_current;
extern int menu_current_size;

extern int menu_horizontal_scroll; // For scrolling long texts that don't fit in LCD_WIDTH
extern int menu_cursor_position;   // Tells which menu item is selected
extern int menu_screen_position;   // Tells which part of the menu to render on the screen

enum MENU_MODE
{
  MENU_NORMAL = 0,
  MENU_EDITING_INTEGER,
  MENU_EDITING_BOOL,
  MENU_EDITING_FLOAT,
  MENU_EDITING_TEXT
};

extern enum MENU_MODE menu_mode;

// #####################################################################

extern int menu_original_integer;
extern int menu_integer_position;
extern struct MENU_ENTRY menu_edit_integer[];

extern bool menu_original_bool;
extern struct MENU_ENTRY menu_edit_bool[];

extern float menu_original_float;
extern int menu_float_position;
extern struct MENU_ENTRY menu_edit_float[];

extern char menu_original_text[256];
extern int menu_text_position;
extern int menu_text_scroll;   // For long text that doesn't fit in LCD_WIDTH
extern struct MENU_ENTRY menu_edit_text[];

extern const PROGMEM struct MENU_ENTRY menu_main[];

extern struct MENU_ENTRY menu_messagebox[];
void menu_display_messagebox( );

extern unsigned long menu_routine_millis;

void menu_routine( );

// #####################################################################