// #####################################################################

// Real Time Clock variables

#include "ntp.h"
#include "wifi.h"
#include "menu.h"
#include "led.h"

RTCZero rtc;

int rtc_timezone;          //change this to adapt it to your time zone
bool b_rtc_auto_sync;

bool b_rtc_sync = false;    // is rtc synchronized with ntp server?

const PROGMEM struct MENU_INTEGER_EXTRA extra_timezone =
{
  .on_change = NULL,
  { .units = "h" },
  .minimum = -24,
  .maximum = 24,
  .increment = 1,
  .increment_fast = 5
};

const PROGMEM struct MENU_ENTRY menu_rtc[] =
{
  { { .text = "Auto-Sync" }, .type = MENU_BOOL, .data = (void *) &b_rtc_auto_sync, .user_data = 0, .extra = NULL },
  { { .text = "Synchronized" }, .type = MENU_BOOL, .data = (void *) &b_rtc_sync, .user_data = 0, .extra = NULL },
  { { .text = "Timezone" }, .type = MENU_INTEGER, .data = (void *) &rtc_timezone, .user_data = 0, .extra = (void *) &extra_timezone },
  { { .text = "" }, .type = MENU_NOTHING }
};

extern unsigned long current_millis;
unsigned long ntp_routine_millis = 5000;
bool b_rtc_initialized = false;

// #####################################################################

void ntp_routine( )
{
  // check timer
  if( current_millis < ntp_routine_millis )
    return;
  ntp_routine_millis = current_millis + 5000;

  if( b_rtc_initialized == false )
  {
    // RTC initialization
    rtc.begin( );
    Serial.println(F("[RTC] RTC initialized."));

    b_rtc_initialized = true;
  }

  if( b_rtc_initialized == true && wifi_status == WL_CONNECTED && b_rtc_sync == false && b_rtc_auto_sync == true )
  {
    // We have connection to WiFi, but no time synchronization
    time_t epoch = (time_t) WiFi.getTime();

    // Check if NTP request was fullfilled
    if( epoch == 0 )
    {
      led_set_state( LED_TIMESYNC_PIN, LED_BLINK );
      Serial.println( F("[RTC] NTP unreachable.") );
    }
    else
    {
      // Update RTC
      Serial.print( F("[RTC] Epoch received: ") );
      Serial.println( epoch );
      rtc.setEpoch( epoch );
      b_rtc_sync = true;

      // Print Date/Time
      epoch += rtc_timezone * 60 * 60; // timezone correction
      struct tm* tmp = gmtime( &epoch ); // struct tm is a structure that contains a calendar time broken down into its components
      char txt_buffer[128];
      strftime( txt_buffer, sizeof(txt_buffer), "%d.%m.%Y %H:%M:%S", tmp );
      Serial.print( F("[RTC] Current time: ") );
      Serial.println( txt_buffer );
    }
  }

  // Update LED status
  if( b_rtc_sync == true )
    led_set_state( LED_TIMESYNC_PIN, LED_ON );
  else
    led_set_state( LED_TIMESYNC_PIN, LED_OFF );
}

// #####################################################################
