// #####################################################################

// Power Management Integrated Circuit subsystem

#include "pmic.h"
#include "menu.h"
#include "led.h"

#include <Arduino_PMIC.h>

#define USB_PIN   15

float pmic_input_current_limit;
float pmic_input_voltageLimit;
float pmic_minimum_system_voltage;
float pmic_charge_voltage;
float pmic_charge_current;

float battery_voltage = NAN;
float usb_voltage = NAN;

char pmic_volts_txt[16];
char pmic_usb_voltage_txt[16];

const PROGMEM struct MENU_FLOAT_EXTRA extra_inputcurrentlimit =
{
  .on_change = NULL,
  { .units = "A" },
  .precision = 2,
  .f_minimum = 0.5,
  .f_maximum = 2.0,
  .f_increment = 1.0,
  .f_increment_fast = 5.0
};

const PROGMEM struct MENU_FLOAT_EXTRA extra_inputvoltlimit =
{
  .on_change = NULL,
  { .units = "V" },
  .precision = 2,
  .f_minimum = 3.88,
  .f_maximum = 4.0,
  .f_increment = 1.0,
  .f_increment_fast = 5.0
};

const PROGMEM struct MENU_FLOAT_EXTRA extra_minsysvoltage =
{
  .on_change = NULL,
  { .units = "V" },
  .precision = 2,
  .f_minimum = 3.5,
  .f_maximum = 4.0,
  .f_increment = 1.0,
  .f_increment_fast = 5.0
};

const PROGMEM struct MENU_FLOAT_EXTRA extra_chargevoltage =
{
  .on_change = NULL,
  { .units = "V" },
  .precision = 2,
  .f_minimum = 3.7,
  .f_maximum = 4.2,
  .f_increment = 1.0,
  .f_increment_fast = 5.0
};

const PROGMEM struct MENU_FLOAT_EXTRA extra_chargecurrent =
{
  .on_change = NULL,
  { .units = "A" },
  .precision = 2,
  .f_minimum = 0.1,
  .f_maximum = 0.5,
  .f_increment = 1.0,
  .f_increment_fast = 5.0
};

struct MENU_ENTRY menu_battery[] =
{          //  ################# - 17 char limit for 20x4 LCD
  { { .text = "Volt: " }, .type = MENU_LABEL, .data = (void *) &pmic_volts_txt, .user_data = 0, .extra = NULL },
  { { .text = "Status: " }, .type = MENU_LABEL, .data = (void *) "N/A", .user_data = 0, .extra = NULL },
  { { .text = "USBMode: " }, .type = MENU_LABEL, .data = (void *) "N/A", .user_data = 0, .extra = NULL },
  { { .text = "USBVolts: " }, .type = MENU_LABEL, .data = (void *) &pmic_usb_voltage_txt, .user_data = 0, .extra = NULL },
  { { .text = "ChargeFault: " }, .type = MENU_LABEL, .data = (void *) "N/A", .user_data = 0, .extra = NULL },
  { { .text = "TempFault: " }, .type = MENU_LABEL, .data = (void *) "N/A", .user_data = 0, .extra = NULL },
  { { .text = "BatOverVolt: " }, .type = MENU_LABEL, .data = (void *) "N/A", .user_data = 0, .extra = NULL },
  { { .text = "Force Charge" }, .type = MENU_COMMAND, .data = (void *) pmic_force_charge, .user_data = 0, .extra = NULL },
  { { .text = "InputCurrentLimit" }, .type = MENU_FLOAT, .data = (void *) &pmic_input_current_limit, .user_data = 0, .extra = (void *) &extra_inputcurrentlimit },
  { { .text = "InputVoltLimit" }, .type = MENU_FLOAT, .data = (void *) &pmic_input_voltageLimit, .user_data = 0, .extra = (void *) &extra_inputvoltlimit },
  { { .text = "MinSysVoltage" }, .type = MENU_FLOAT, .data = (void *) &pmic_minimum_system_voltage, .user_data = 0, .extra = (void *) &extra_minsysvoltage },
  { { .text = "ChargeVoltage" }, .type = MENU_FLOAT, .data = (void *) &pmic_charge_voltage, .user_data = 0, .extra = (void *) &extra_chargevoltage },
  { { .text = "ChargeCurrent" }, .type = MENU_FLOAT, .data = (void *) &pmic_charge_current, .user_data = 0, .extra = (void *) &extra_chargecurrent },
  { { .text = "" }, .type = MENU_NOTHING }
};

extern unsigned long current_millis;
bool b_pmic_initialized = false;
unsigned long pmic_routine_millis = 10000;

// #####################################################################

void pmic_force_charge( struct MENU_ENTRY *menu_entry )
{
  PMIC.enableCharge( );
}

// #####################################################################

void pmic_routine( )
{
  // check timer
  if( current_millis < pmic_routine_millis )
    return;
  pmic_routine_millis = current_millis + 5000;

  // Update labels in the menu
  // TODO: Add "Error" and "Discharging" state
  switch( PMIC.chargeStatus( ) )
  {
    default:
    case NOT_CHARGING: menu_battery[1].data = (void *) "NOT_CHARGING"; break;
    case PRE_CHARGING: menu_battery[1].data = (void *) "PRE_CHARGING"; led_set_state( LED_POWER_PIN, LED_ON ); break;
    case FAST_CHARGING: menu_battery[1].data = (void *) "FAST_CHARGING"; led_set_state( LED_POWER_PIN, LED_ON ); break;
    case CHARGE_TERMINATION_DONE: menu_battery[1].data = (void *) "CHARGE_TERMINATION_DONE"; break;
  }

  // getChargeFault() returns charge fault status
  switch( PMIC.getChargeFault( ) )
  {
    case INPUT_OVER_VOLTAGE: menu_battery[4].data = (void *) "Input over voltage fault occurred"; break;
    case THERMAL_SHUTDOWN: menu_battery[4].data = (void *) "Thermal shutdown occurred"; break;
    case CHARGE_SAFETY_TIME_EXPIRED: menu_battery[4].data = (void *) "Charge safetly timer expired"; break;
    case NO_CHARGE_FAULT: menu_battery[4].data = (void *) "No Charge fault"; break;
    default: break;
  }

  switch( PMIC.hasBatteryTemperatureFault( ) )
  {
    case NO_TEMPERATURE_FAULT: menu_battery[5].data = (void *) "No temperature fault"; break;
    case LOWER_THRESHOLD_TEMPERATURE_FAULT: menu_battery[5].data = (void *) "Lower threshold Battery temperature fault"; break;
    case HIGHER_THRESHOLD_TEMPERATURE_FAULT: menu_battery[5].data = (void *) "Higher threshold Battery temperature fault"; break;
    default: break;
  }

  // The isBatteryInOverVoltage() returns if battery over-voltage fault occurs.
  // When battery over voltage occurs, the charger device immediately disables
  // charge and sets the battery fault bit, in fault register, to high.
  if( PMIC.isBatteryInOverVoltage( ) )
    menu_battery[6].data = (void *) "Error: battery over voltage fault";
  else
    menu_battery[6].data = (void *) "OK";

  // Read voltages
  battery_voltage = analogRead( ADC_BATTERY ) * ( 4.3 / 1023.0 );

  //
  // Vin   O----\
  //            |                          R2
  //           +-+               Vout = --------- * Vin
  //           | |  R1                   R1 + R2
  //           +-+
  //            |
  //            *------O Vout                  R1 + R2
  //            |                Vin = Vout * ---------
  //           +-+                               R2
  //           | |  R2
  //           +-+
  //            |
  // GND  O-----/
  //
  const float USB_R1 = 11.86; // Measured resistance R1, kOhm
  const float USB_R2 = 11.85; // Measured resistance R2, kOhm
  usb_voltage = ( ((float) analogRead( USB_PIN )) * (3.3 / 1024.0) ) * (USB_R1 + USB_R2) / USB_R2;

  snprintf( pmic_volts_txt, sizeof(pmic_volts_txt), "%4.2f V", battery_voltage );
  snprintf( pmic_usb_voltage_txt, sizeof(pmic_usb_voltage_txt), "%4.2f V", usb_voltage );
  
  switch( PMIC.USBmode( ) )
  {
    default:
    case UNKNOWN_MODE: menu_battery[2].data = (void *) "UNKNOWN_MODE"; break;
    case USB_HOST_MODE: menu_battery[2].data = (void *) "USB_HOST_MODE"; break;
    case ADAPTER_PORT_MODE: menu_battery[2].data = (void *) "ADAPTER_PORT_MODE"; break;
    case BOOST_MODE: menu_battery[2].data = (void *) "BOOST_MODE"; break;
  }

  if( b_pmic_initialized == false )
  {
    pinMode( USB_PIN, INPUT );
    
    if( battery_voltage < 1.0 )
    {
      // Battery dead or not connected
      // Do not start trying to charge it
      led_set_state( LED_POWER_PIN, LED_BLINK );
      Serial.println(F("[PMIC] Battery dead or missing."));
      return;
    }

    if( !PMIC.begin( ) )
    {
      led_set_state( LED_POWER_PIN, LED_BLINK );
      Serial.println(F("[PMIC] Failed to initialize PMIC."));
      return;
    }

    // Set the input current limit to 2 A and the overload input voltage to 3.88 V
    if( !PMIC.setInputCurrentLimit( pmic_input_current_limit ) )
    {
      led_set_state( LED_POWER_PIN, LED_BLINK );
      Serial.println(F("[PMIC] Error in set input current limit."));
      return;
    }
  
    if( !PMIC.setInputVoltageLimit( pmic_input_voltageLimit ) )
    {
      led_set_state( LED_POWER_PIN, LED_BLINK );
      Serial.println(F("[PMIC] Error in set input voltage limit."));
      return;
    }
  
    // set the minimum voltage used to feeding the module embed on Board
    if( !PMIC.setMinimumSystemVoltage( pmic_minimum_system_voltage ) )
    {
      led_set_state( LED_POWER_PIN, LED_BLINK );
      Serial.println(F("[PMIC] Error in set minimum system volage."));
      return;
    }
  
    // Set the desired charge voltage to 4.11 V
    if( !PMIC.setChargeVoltage( pmic_charge_voltage ) )
    {
      led_set_state( LED_POWER_PIN, LED_BLINK );
      Serial.println(F("[PMIC] Error in set charge volage."));
      return;
    }
  
    // Set the charge current to 375 mA
    // the charge current should be definde as maximum at (C for hour)/2h
    // to avoid battery explosion (for example for a 750mAh battery set to 0.375 A)
    if( !PMIC.setChargeCurrent( pmic_charge_current ) )
    {
      led_set_state( LED_POWER_PIN, LED_BLINK );
      Serial.println(F("[PMIC] Error in set charge current."));
      return;
    }
    
    Serial.println(F("[PMIC] Power Management Integrated Circuit initialized."));
    b_pmic_initialized = true;
  }
  else
  {
    if( battery_voltage < 1.0 )
    {
      // Battery dead or not connected
      // Do not start trying to charge it
      led_set_state( LED_POWER_PIN, LED_BLINK );
      Serial.println(F("[PMIC] Battery dead or missing."));
      return;
    }
    
    // Detect presence of USB voltage
    // Somehow battery voltage leaks to USB voltage, when USB is disconnected
    // Battery Voltage = 4.09 V
    // USB Voltage = 3.29 V
    if( usb_voltage > 3.5 )
    {
      // Enable the Charger
      bool charge_result = PMIC.enableCharge( );
      if( PMIC.chargeStatus( ) != CHARGE_TERMINATION_DONE && PMIC.chargeStatus( ) != FAST_CHARGING )
      {
        if( !charge_result )
        {
          led_set_state( LED_POWER_PIN, LED_BLINK );
          Serial.println(F("[PMIC] Error enabling Charge mode."));
        }
        else
        {
          led_set_state( LED_POWER_PIN, LED_ON );
          Serial.println(F("[PMIC] Started charging the battery."));
        }
      }
    }

    // Disable the charger
    if( PMIC.chargeStatus( ) == NOT_CHARGING || PMIC.chargeStatus( ) == CHARGE_TERMINATION_DONE )
    {
      led_set_state( LED_POWER_PIN, LED_OFF );
      Serial.println(F("[PMIC] Battery fully charged."));
    }

    // canRunOnBattery() returns true if the battery voltage is < the minimum
    // systems voltage
    if( PMIC.canRunOnBattery( ) )
    {
      // loop until charge is done
      if( PMIC.chargeStatus( ) != CHARGE_TERMINATION_DONE )
      {
        return;
      }

      // Disable the charger
      if( PMIC.chargeStatus( ) != NOT_CHARGING )
      {
        Serial.println( F("[PMIC] Disable Charge mode.") );
        if( !PMIC.disableCharge( ) )
        {
          led_set_state( LED_POWER_PIN, LED_BLINK );
          Serial.println(F("[PMIC] Error disabling Charge mode."));
        }
      }
    }
  }
}

// #####################################################################
