// #####################################################################

// Sensor variables

#include "menu.h"

#include <Arduino_MKRENV.h>

float env_temperature = NAN;
float env_humidity = NAN;
float env_pressure = NAN;
float env_illuminance = NAN;
float env_uva = NAN;
float env_uvb = NAN;
float env_uvIndex = NAN;

int sensor_attempts = 0;
int sensor_max_attempts;

const PROGMEM struct MENU_INTEGER_EXTRA extra_maxattempts =
{
  .on_change = NULL,
  { .units = "" },
  .minimum = 0,
  .maximum = 3,
  .increment = 1,
  .increment_fast = 5
};

struct MENU_ENTRY menu_env_shield[] =
{
  { { .text = "MaxAttempts" }, .type = MENU_INTEGER, .data = (void *) &sensor_max_attempts, .user_data = 0, .extra = (void *) &extra_maxattempts },
  { { .text = "Status: " }, .type = MENU_LABEL, .data = (void *) "N/A", .user_data = 0, .extra = NULL },
  { { .text = "" }, .type = MENU_NOTHING }
};

extern unsigned long current_millis;
bool b_sensor_initialized = false;
unsigned long sensor_routine_millis = 1000;

// #####################################################################

void sensor_routine( )
{
  // check timer
  if( current_millis < sensor_routine_millis )
    return;
  sensor_routine_millis = current_millis + 1000;

  if( b_sensor_initialized == false )
  {
    // If ENV Shield fails to initialize 4 times, then on 5th time
    // it will crash arduino
    if( sensor_attempts < sensor_max_attempts )
    {
      int result = ENV.begin( );
      if( result == 1 )
      {
        menu_env_shield[1].data = (void *) "OK";
        Serial.println( F( "[Sensor] MKR ENV Shield initialized succesfully." ) );
        b_sensor_initialized = true;
      }
      else 
      {
        menu_env_shield[1].data = (void *) "FAIL";
        sensor_attempts++;
        Serial.println( F( "[Sensor] Failed to initialize MKR ENV Shield!" ) );
        return;
      }
    }
  }
  else
  {
    // Read all the sensor values
    env_temperature = ENV.readTemperature( );
    env_humidity    = ENV.readHumidity( );
    env_pressure    = ENV.readPressure( );
    env_illuminance = ENV.readIlluminance( );
    env_uva         = ENV.readUVA( );
    env_uvb         = ENV.readUVB( );
    env_uvIndex     = ENV.readUVIndex( );
  }
}

// #####################################################################
