// #####################################################################

// WiFi settings

#include "wifi.h"
#include "menu.h"
#include "keyboard.h"
#include "led.h"
#include "watchdog.h"
#include "webserver.h"

#include "arduino_secrets.h"
///////please enter your sensitive data in the Secret tab/arduino_secrets.h
char ssid[] = SECRET_SSID;    // your network SSID (name)
char pass[] = SECRET_PASS;    // your network password (use for WPA, or use as key for WEP)
int keyIndex = 0;             // your network key Index number (needed only for WEP)

char wifi_firmware[8];
char wifi_mac[20];
char wifi_local_ip[24] = "0.0.0.0";
char wifi_connected_rssi[16] = "N/A dBm";

int prev_wifi_status = WL_IDLE_STATUS;
int wifi_status = WL_IDLE_STATUS;
int wifi_numSsid; // Number of networks found after scan
bool b_wifi_autoconnect = false;

void wifi_edit_net( struct MENU_ENTRY *menu_entry );
void wifi_connect_net( struct MENU_ENTRY *menu_entry );
void wifi_delete_net( struct MENU_ENTRY *menu_entry );
void wifi_add_net( struct MENU_ENTRY *menu_entry );
void wifi_view_net( struct MENU_ENTRY *menu_entry );
void wifi_start_scan( struct MENU_ENTRY *menu_entry );
void wifi_disconnect( struct MENU_ENTRY *menu_entry );
void wifi_net_on_open( struct MENU_ENTRY *menu_entry );

struct WIFI_NETWORK wifi_networks[WIFI_SAVED_NETWORKS_MAX];

struct WIFI_SCAN
{
  char ssid[WIFI_MAX_SSID_LENGTH+1];
  int rssi;             // Signal strength, dBm
  int channel;
  byte bssid[6];
  int encryption_type;  // WEP/WPA/WPA2
}
wifi_scan[10];

struct MENU_ENTRY menu_wifi_networks[] =
{
  { { .text = "Networks:" }, .type = MENU_NOOP },
  { { .text = "Net0" }, .type = MENU_COMMAND, .data = (void *) wifi_edit_net, .user_data = 0, .extra = NULL },
  { { .text = "Net1" }, .type = MENU_COMMAND, .data = (void *) wifi_edit_net, .user_data = 1, .extra = NULL },
  { { .text = "Net2" }, .type = MENU_COMMAND, .data = (void *) wifi_edit_net, .user_data = 2, .extra = NULL },
  { { .text = "Net3" }, .type = MENU_COMMAND, .data = (void *) wifi_edit_net, .user_data = 3, .extra = NULL },
  { { .text = "Net4" }, .type = MENU_COMMAND, .data = (void *) wifi_edit_net, .user_data = 4, .extra = NULL },
  { { .text = "Net5" }, .type = MENU_COMMAND, .data = (void *) wifi_edit_net, .user_data = 5, .extra = NULL },
  { { .text = "Net6" }, .type = MENU_COMMAND, .data = (void *) wifi_edit_net, .user_data = 6, .extra = NULL },
  { { .text = "Net7" }, .type = MENU_COMMAND, .data = (void *) wifi_edit_net, .user_data = 7, .extra = NULL },
  { { .text = "Net8" }, .type = MENU_COMMAND, .data = (void *) wifi_edit_net, .user_data = 8, .extra = NULL },
  { { .text = "Net9" }, .type = MENU_COMMAND, .data = (void *) wifi_edit_net, .user_data = 9, .extra = NULL },
  { { .text = "" }, .type = MENU_NOTHING }
};

const PROGMEM struct MENU_TEXT_EXTRA extra_ssid =
{
  .text_size = WIFI_MAX_SSID_LENGTH
};

const PROGMEM struct MENU_TEXT_EXTRA extra_pass =
{
  .text_size = WIFI_MAX_PASS_LENGTH
};

// Menu for editing saved networks
char wifi_edit_net_txt[8];
struct MENU_ENTRY menu_edit_net[] =
{
  { { .text = "Net: " }, .type = MENU_LABEL, .data = (void *) &wifi_edit_net_txt, .user_data = 0, .extra = NULL },
  { { .text = "SSID" }, .type = MENU_TEXT, .data = NULL, .user_data = 0, .extra = (void *) &extra_ssid },
  { { .text = "PASS" }, .type = MENU_TEXT, .data = NULL, .user_data = 0, .extra = (void *) &extra_pass },
  { { .text = "Connect" }, .type = MENU_COMMAND, .data = (void *) wifi_connect_net, .user_data = 0, .extra = NULL },
  { { .text = "Delete" }, .type = MENU_COMMAND, .data = (void *) wifi_delete_net, .user_data = 0, .extra = NULL },
  { { .text = "" }, .type = MENU_NOTHING }
};

// Menu for viewing scanned networks
char wifi_rssi_text[16];
char wifi_channel_text[8];
char wifi_bssid_text[20];
struct MENU_ENTRY menu_view_net[] =
{
  { { .text = "SSID: " }, .type = MENU_LABEL, .data = NULL, .user_data = 0, .extra = NULL },
  { { .text = "RSSI: " }, .type = MENU_LABEL, .data = (void *) &wifi_rssi_text[0], .user_data = 0, .extra = NULL },
  { { .text = "Channel: " }, .type = MENU_LABEL, .data = (void *) &wifi_channel_text[0], .user_data = 0, .extra = NULL },
  { { .text = "BSSID: " }, .type = MENU_LABEL, .data = (void *) &wifi_bssid_text[0], .user_data = 0, .extra = NULL },
  { { .text = "Encryption: " }, .type = MENU_LABEL, .data = (void *) "N/A", .user_data = 0, .extra = NULL },
  { { .text = "Add to list" }, .type = MENU_COMMAND, .data = (void *) wifi_add_net, .user_data = 0, .extra = NULL },
  { { .text = "" }, .type = MENU_NOTHING }
};

struct MENU_ENTRY menu_wifi_scan[] =
{
  { { .text = "Start scan" }, .type = MENU_COMMAND, .data = (void *) wifi_start_scan, .user_data = 0, .extra = NULL },
  { { .text = "" }, .type = MENU_COMMAND, .data = (void *) wifi_view_net, .user_data = 0, .extra = NULL },
  { { .text = "" }, .type = MENU_COMMAND, .data = (void *) wifi_view_net, .user_data = 1, .extra = NULL },
  { { .text = "" }, .type = MENU_COMMAND, .data = (void *) wifi_view_net, .user_data = 2, .extra = NULL },
  { { .text = "" }, .type = MENU_COMMAND, .data = (void *) wifi_view_net, .user_data = 3, .extra = NULL },
  { { .text = "" }, .type = MENU_COMMAND, .data = (void *) wifi_view_net, .user_data = 4, .extra = NULL },
  { { .text = "" }, .type = MENU_COMMAND, .data = (void *) wifi_view_net, .user_data = 5, .extra = NULL },
  { { .text = "" }, .type = MENU_COMMAND, .data = (void *) wifi_view_net, .user_data = 6, .extra = NULL },
  { { .text = "" }, .type = MENU_COMMAND, .data = (void *) wifi_view_net, .user_data = 7, .extra = NULL },
  { { .text = "" }, .type = MENU_COMMAND, .data = (void *) wifi_view_net, .user_data = 8, .extra = NULL },
  { { .text = "" }, .type = MENU_COMMAND, .data = (void *) wifi_view_net, .user_data = 9, .extra = NULL },
  { { .text = "" }, .type = MENU_NOTHING }
};

const PROGMEM struct MENU_SUBMENU_EXTRA extra_networks =
{
  .on_open = wifi_net_on_open
};

struct MENU_ENTRY menu_wifi[] =
{
  { { .text = "[Networks]" }, .type = MENU_SUBMENU, .data = &menu_wifi_networks, .user_data = 0, .extra = (void *) &extra_networks  },
  { { .text = "[Scan]" }, .type = MENU_SUBMENU, .data = &menu_wifi_scan, .user_data = 0, .extra = NULL },
  { { .text = "Auto-connect" }, .type = MENU_BOOL, .data = &b_wifi_autoconnect, .user_data = 0, .extra = NULL },
  { { .text = "Disconnect" }, .type = MENU_COMMAND, .data = (void *) wifi_disconnect, .user_data = 0, .extra = NULL },
  { { .text = "Status: " }, .type = MENU_LABEL, .data = (void *) "N/A", .user_data = 0, .extra = NULL },
  { { .text = "Firmware: " }, .type = MENU_LABEL, .data = (void *) "N/A", .user_data = 0, .extra = NULL },
  { { .text = "SSID: " }, .type = MENU_LABEL, .data = (void *) "N/A", .user_data = 0, .extra = NULL },
  { { .text = "MAC: " }, .type = MENU_LABEL, .data = (void *) "N/A", .user_data = 0, .extra = NULL },
  { { .text = "IP: " }, .type = MENU_LABEL, .data = &wifi_local_ip, .user_data = 0, .extra = NULL },
  { { .text = "RSSI: " }, .type = MENU_LABEL, .data = &wifi_connected_rssi, .user_data = 0, .extra = NULL },
  { { .text = "" }, .type = MENU_NOTHING }
};

extern unsigned long current_millis;
unsigned long wifi_routine_millis = 15000;
unsigned long wifi_connect_millis = 20000;
bool b_wifi_initialized = false;

// #####################################################################

// Result: 4C:EB:D6:4C:76:94
void printMacAddress( char *txt_buf, size_t txt_len, byte mac[] )
{
  String txt;
  for (int i = 5; i >= 0; i--)
  {
    if (mac[i] < 16)
    {
      txt.concat( "0" );
    }
    
    txt.concat( String(mac[i], HEX) );
    
    if (i > 0)
    {
      txt.concat( ":" );
    }
  }

  txt.toCharArray( txt_buf, txt_len );
}

// #####################################################################

void wifi_edit_net( struct MENU_ENTRY *menu_entry )
{
  if( menu_entry->user_data < WIFI_SAVED_NETWORKS_MAX )
  {
    // Copy wifi network data to editing menu
    snprintf( wifi_edit_net_txt, sizeof(wifi_edit_net_txt), "%i", menu_entry->user_data );
  
    menu_edit_net[1].data = (void *) &wifi_networks[menu_entry->user_data].ssid;
    menu_edit_net[2].data = (void *) &wifi_networks[menu_entry->user_data].pass;
    menu_edit_net[3].user_data = menu_entry->user_data;
    menu_edit_net[4].user_data = menu_entry->user_data;
  
    keyboard_push_stack( );
    menu_cursor_position = 3;   // Make cursor point at "Connect" menu item
    menu_screen_position = 0;
    menu_current = menu_edit_net;
    menu_current_size = 0;
  }
}

// #####################################################################

void wifi_connect_net( struct MENU_ENTRY *menu_entry )
{
  if( b_wifi_initialized == false )
  {
    //                                 ###########_##### - 17 char limit for 20x4 LCD (12 char limit first line)
    strncpy( menu_messagebox[0].text, "- Error -", LCD_WIDTH + 1 );
    strncpy( menu_messagebox[1].text, "WiFi not initia-", LCD_WIDTH + 1 );
    strncpy( menu_messagebox[2].text, "lized", LCD_WIDTH + 1 );
    menu_display_messagebox( );
    return;
  }

  Serial.print( F("[WiFi] Attempting to connect to SSID: ") );
  Serial.println( wifi_networks[menu_entry->user_data].ssid );

  // Connect to WPA/WPA2 network. Change this line if using open or WEP network:
  wifi_status = WiFi.begin(
    wifi_networks[menu_entry->user_data].ssid,
    wifi_networks[menu_entry->user_data].pass
  );

  keyboard_pop_stack( );
  keyboard_pop_stack( );

  wifi_connect_millis = current_millis + 10000;
}

// #####################################################################

void wifi_delete_net( struct MENU_ENTRY *menu_entry )
{
  wifi_networks[menu_entry->user_data].ssid[0] = 0;
  wifi_networks[menu_entry->user_data].pass[0] = 0;

  //                                 ###########_##### - 17 char limit for 20x4 LCD (12 char limit first line)
  strncpy( menu_messagebox[0].text, "- Message -", LCD_WIDTH + 1 );
  strncpy( menu_messagebox[1].text, "WiFi slot cleared", LCD_WIDTH + 1 );
  strncpy( menu_messagebox[2].text, "", LCD_WIDTH + 1 );
  menu_display_messagebox( );
}

// #####################################################################

void wifi_add_net( struct MENU_ENTRY *menu_entry )
{
  // Find empty slot
  for( int i = 0; i < WIFI_SAVED_NETWORKS_MAX; i++ )
  {
    if( wifi_networks[i].ssid[0] == 0 && wifi_networks[i].pass[0] == 0 )
    {
      strncpy( wifi_networks[i].ssid, wifi_scan[menu_entry->user_data].ssid, WIFI_MAX_SSID_LENGTH );

      //                                 ###########_##### - 17 char limit for 20x4 LCD (12 char limit first line)
      strncpy( menu_messagebox[0].text, "- Message -", LCD_WIDTH + 1 );
      strncpy( menu_messagebox[1].text, "Stored to WiFi", LCD_WIDTH + 1 );
      snprintf( menu_messagebox[2].text, LCD_WIDTH - 3, "slot: Net%i", i );
      menu_display_messagebox( );

      return;
    }
  }

  //                                 ###########_##### - 17 char limit for 20x4 LCD (12 char limit first line)
  strncpy( menu_messagebox[0].text, "- Error -", LCD_WIDTH + 1 );
  strncpy( menu_messagebox[1].text, "WiFi slot not", LCD_WIDTH + 1 );
  strncpy( menu_messagebox[2].text, "found", LCD_WIDTH + 1 );
  menu_display_messagebox( );
}

// #####################################################################

void wifi_view_net( struct MENU_ENTRY *menu_entry )
{
  if( menu_entry->user_data < WIFI_SAVED_NETWORKS_MAX )
  {
    // Copy wifi network data to viewing menu
    menu_view_net[0].data = (void *) &wifi_scan[menu_entry->user_data].ssid[0];
    snprintf( wifi_rssi_text, sizeof(wifi_rssi_text), "%i dBm", wifi_scan[menu_entry->user_data].rssi );
    snprintf( wifi_channel_text, sizeof(wifi_channel_text), "%i", wifi_scan[menu_entry->user_data].channel );
    printMacAddress( wifi_bssid_text, sizeof(wifi_bssid_text), &wifi_scan[menu_entry->user_data].bssid[0] );
    switch( wifi_scan[menu_entry->user_data].encryption_type )
    {
      case ENC_TYPE_WEP:  menu_view_net[4].data = (void *) "WEP"; break;
      case ENC_TYPE_TKIP: menu_view_net[4].data = (void *) "WPA"; break;
      case ENC_TYPE_CCMP: menu_view_net[4].data = (void *) "WPA2"; break;
      case ENC_TYPE_NONE: menu_view_net[4].data = (void *) "None"; break;
      case ENC_TYPE_AUTO: menu_view_net[4].data = (void *) "Auto"; break;
      case ENC_TYPE_UNKNOWN:
      default:            menu_view_net[4].data = (void *) "Unknown"; break;
    }
    menu_view_net[5].user_data = menu_entry->user_data;

    keyboard_push_stack( );
    menu_cursor_position = 0;
    menu_screen_position = 0;
    menu_current = menu_view_net;
    menu_current_size = 0;
  }
}

// #####################################################################

void wifi_start_scan( struct MENU_ENTRY *menu_entry )
{
  int max_numSsid;

  // Display message that scanning process is started
  strcpy( menu_entry->text, "Scanning..." );

  // Clear the list to visually see it's going to be a new list
  for( int i = 1; i < sizeof(wifi_scan) / sizeof(struct MENU_ENTRY) - 1; i++ )
    menu_wifi_scan[i].text[0] = 0;
  memset( wifi_scan, 0, sizeof(wifi_scan) );

  // Force render with a dirty hack
  menu_routine_millis = 0;
  menu_routine( );
  lcd_routine_millis = 0;
  lcd_routine( );

  // Start network scan
  int numSsid = WiFi.scanNetworks( );
  if( numSsid == -1 )
  {
    //                                 ###########_##### - 17 char limit for 20x4 LCD (12 char limit first line)
    strncpy( menu_messagebox[0].text, "- Error -", LCD_WIDTH + 1 );
    strncpy( menu_messagebox[1].text, "WiFi networks not", LCD_WIDTH + 1 );
    strncpy( menu_messagebox[2].text, "found", LCD_WIDTH + 1 );
    menu_display_messagebox( );
    goto fail;
  }

  // Limit number of networks to the size of the array
  max_numSsid = sizeof(wifi_scan) / sizeof(struct WIFI_SCAN);
  if( numSsid > max_numSsid )
  {
    numSsid = max_numSsid;

    //                                 ###########_##### - 17 char limit for 20x4 LCD (12 char limit first line)
    strncpy( menu_messagebox[0].text, "- Warning -", LCD_WIDTH + 1 );
    strncpy( menu_messagebox[1].text, "WiFi network buf-", LCD_WIDTH + 1 );
    strncpy( menu_messagebox[2].text, "fer overflow", LCD_WIDTH + 1 );
    menu_display_messagebox( );
  }

  // Copy information about wifi networks around into local array
  memset( wifi_scan, 0, sizeof(wifi_scan) );
  for( int i = 0; i < numSsid; i++ )
  {
    strncpy( wifi_scan[i].ssid, WiFi.SSID( i ), WIFI_MAX_SSID_LENGTH );
    strncpy( menu_wifi_scan[i + 1].text, WiFi.SSID( i ), LCD_WIDTH - 3 );
    wifi_scan[i].rssi = WiFi.RSSI( i );
    wifi_scan[i].channel = WiFi.channel( i );
    byte bssid[6]; WiFi.BSSID( i, bssid );
    memcpy( &wifi_scan[i].bssid, bssid, sizeof(bssid) );
    wifi_scan[i].encryption_type = WiFi.encryptionType( i );
  }

fail:
  // Display message that scanning process is ended
  strcpy( menu_entry->text, "Start scan" );  
}

void wifi_disconnect( struct MENU_ENTRY *menu_entry )
{
  WiFi.end();

  //                                 ###########_##### - 17 char limit for 20x4 LCD (12 char limit first line)
  strncpy( menu_messagebox[0].text, "- Message -", LCD_WIDTH + 1 );
  strncpy( menu_messagebox[1].text, "WiFi turned off", LCD_WIDTH + 1 );
  strncpy( menu_messagebox[2].text, "", LCD_WIDTH + 1 );
  menu_display_messagebox( );
}

// #####################################################################

void wifi_net_on_open( struct MENU_ENTRY *menu_entry )
{
  //
  // Update network names in the menu_wifi_networks when user opens list of saved networks
  //

  for( int i = 0; i < WIFI_SAVED_NETWORKS_MAX; i++ )
  {
    snprintf( menu_wifi_networks[i + 1].text, LCD_WIDTH - 3, "Net%i: %s", i, wifi_networks[i].ssid );
  }
}

// #####################################################################

void wifi_routine()
{
  // check timer
  if( current_millis < wifi_routine_millis )
    return;
  wifi_routine_millis = current_millis + 1000;

  // get wifi status
  prev_wifi_status = wifi_status;
  wifi_status = WiFi.status();

  switch( wifi_status )
  {
    case WL_CONNECTED:        menu_wifi[4].data = (void *) "WL_CONNECTED: assigned when connected to a WiFi network"; break;
    case WL_AP_CONNECTED:     menu_wifi[4].data = (void *) "WL_AP_CONNECTED : assigned when a device is connected in Access Point mode"; break;
    case WL_AP_LISTENING:     menu_wifi[4].data = (void *) "WL_AP_LISTENING : assigned when the listening for connections in Access Point mode"; break;
    case WL_NO_MODULE:        menu_wifi[4].data = (void *) "WL_NO_MODULE: assigned when the communication with an integrated WiFi module fails"; break;
    case WL_IDLE_STATUS:      menu_wifi[4].data = (void *) "WL_IDLE_STATUS: it is a temporary status assigned when WiFi.begin() is called and remains active until the number of attempts expires (resulting in WL_CONNECT_FAILED) or a connection is established (resulting in WL_CONNECTED)"; break;
    case WL_NO_SSID_AVAIL:    menu_wifi[4].data = (void *) "WL_NO_SSID_AVAIL: assigned when no SSID are available"; break;
    case WL_SCAN_COMPLETED:   menu_wifi[4].data = (void *) "WL_SCAN_COMPLETED: assigned when the scan networks is completed"; break;
    case WL_CONNECT_FAILED:   menu_wifi[4].data = (void *) "WL_CONNECT_FAILED: assigned when the connection fails for all the attempts"; break;
    case WL_CONNECTION_LOST:  menu_wifi[4].data = (void *) "WL_CONNECTION_LOST: assigned when the connection is lost"; break;
    case WL_DISCONNECTED:     menu_wifi[4].data = (void *) "WL_DISCONNECTED: assigned when disconnected from a network;"; break;
  }

  switch( wifi_status )
  {
    case WL_CONNECTED:
    case WL_AP_CONNECTED:     led_set_state( LED_WIFI_PIN, LED_ON ); break;
    case WL_NO_MODULE:
    case WL_NO_SSID_AVAIL:
    case WL_CONNECT_FAILED:
    case WL_DISCONNECTED:     led_set_state( LED_WIFI_PIN, LED_BLINK ); break;
    default:                  led_set_state( LED_WIFI_PIN, LED_OFF ); break;
  }

  // Put local ip address into the menu
  IPAddress ip = WiFi.localIP();    // IPAddress is uint8_t
  snprintf( wifi_local_ip, sizeof(wifi_local_ip), "%i.%i.%i.%i", (int) ip[0], (int) ip[1], (int) ip[2], (int) ip[3] );

  long rssi = WiFi.RSSI( ); // Put received signal strength into the menu
  snprintf( wifi_connected_rssi, sizeof(wifi_connected_rssi), "%li dBm", rssi );

  if( b_wifi_initialized == false )
  {
    // wifi is broken?
    if( wifi_status == WL_NO_SHIELD )
    {
      Serial.println(F("[WiFi] WiFi shield not present."));
      return;
    }
    if( wifi_status == WL_NO_MODULE )
    {
      Serial.println(F("[WiFi] The communication with an integrated WiFi module fails."));
      return;
    }

    // Get firmware vesion of our WiFi module
    String fv = WiFi.firmwareVersion();
    fv.toCharArray( wifi_firmware, sizeof(wifi_firmware) );
    menu_wifi[5].data = (void *) wifi_firmware;

    // Get MAC address of our WiFi module
    byte mac[6];
    WiFi.macAddress(mac);
    printMacAddress( wifi_mac, sizeof(wifi_mac), mac );
    menu_wifi[7].data = (void *) wifi_mac;
    
    b_wifi_initialized = true;
  }
  else
  {
    // catch the moment when connection was lost
    if( prev_wifi_status == WL_CONNECTED && wifi_status == WL_CONNECTION_LOST )
    {
      Serial.println( F("[WiFi] Connection is lost.") );
    }
  
    // try to connect
    if( b_wifi_autoconnect == true && wifi_status != WL_CONNECTED && wifi_connect_millis < current_millis )
    {
      // Connect to WPA/WPA2 network. Change this line if using open or WEP network:
      //wifi_status = WiFi.begin( ssid, pass );

      //
      // Scan for networks and connect to the first found in local list
      //

      if( wifi_status != WL_SCAN_COMPLETED )
      {
        // Start network scan
        Serial.println( F("[WiFi] Scanning networks...") );
        wifi_numSsid = WiFi.scanNetworks( );
        watchdog.clear( );
      }
      else
      {
        if( wifi_numSsid != -1 )
        {
          // Check each network found with our local list and connect to first match
          bool stop = false;
          for( int i = 0; i < wifi_numSsid && !stop; i++ )
          {
            for( int j = 0; j < sizeof(wifi_networks) / sizeof(struct WIFI_NETWORK) && !stop; j++ )
            {
              if( wifi_networks[j].ssid[0] == 0 )
                continue;
  
              if( strcmp( wifi_networks[j].ssid, WiFi.SSID( i ) ) == 0 )
              {
                Serial.print( F("[WiFi] Attempting to connect to SSID: ") );
                Serial.println( wifi_networks[j].ssid );
        
                wifi_status = WiFi.begin( wifi_networks[j].ssid, wifi_networks[j].pass );
                watchdog.clear( );
                if( wifi_status == WL_CONNECTED )
                {
                  stop = true;
                }
                else if( wifi_status == WL_CONNECT_FAILED )
                {
                  Serial.println( F("[WiFi] Connection failed.") );
                }
                break;
              }
            }
          }
        }
      }

      wifi_connect_millis = current_millis + 10000;
    }

    // Update SSID label in the menu after auto-connect attempts
    menu_wifi[6].data = (void *) WiFi.SSID( );
  
    // catch the moment when connection was lost
    if( prev_wifi_status == WL_IDLE_STATUS && wifi_status == WL_CONNECT_FAILED )
    {
      Serial.println( F("[WiFi] Connection failed.") );
    }
  
    // catch the moment when successfully connected to wifi router
    if( prev_wifi_status != WL_CONNECTED && wifi_status == WL_CONNECTED )
    {
      Serial.print( F("[WiFi] Successfully connected to: ") );
      Serial.println( WiFi.SSID( ) );
      // you're connected now, so print out the status:
      printWiFiStatus( );
  
      // Reset webserver initialization
      // seems to be that webserver needs to be reinitializated when wifi reconnects
      // otherwise it doesn't work
      b_webserver_initialized = false;
    }
  }
}

// #####################################################################

void printWiFiStatus( )
{
  // print the SSID of the network you're attached to:
  Serial.print( F("[WiFi]     SSID: ") );
  Serial.println( WiFi.SSID( ) );

  // print your WiFi shield's IP address:
  IPAddress ip = WiFi.localIP( );
  Serial.print( F("[WiFi]     IP Address: ") );
  Serial.println( ip );

  // print the received signal strength:
  long rssi = WiFi.RSSI( );
  Serial.print( F("[WiFi]     Signal strength (RSSI):") );
  Serial.print( rssi );
  Serial.println( F(" dBm") );
}

// #####################################################################
